import csv
from dataclasses import dataclass, field
from datetime import datetime
from functools import partial
import io
import json
from multiprocessing.managers import DictProxy
import signal
import sys
import numpy as np
from pathlib import Path
from xopen import xopen
import polars as pl
from tqdm import tqdm
from typing import Any
import statsmodels.formula.api as smf
from statsmodels.discrete.discrete_model import BinaryResultsWrapper
from statsmodels.genmod.generalized_linear_model import GLMResultsWrapper
from statsmodels.genmod.families.family import Gaussian
from statsmodels.tools.sm_exceptions import (
    ConvergenceWarning,
    PerfectSeparationWarning,
)
import multiprocessing as mp
import warnings

from pyphewas.parser import generate_parser

# We want to treat the Runtime warning like a error so that we can catch it because it generate indicates the perfect separation error
warnings.filterwarnings("error", category=RuntimeWarning)
warnings.filterwarnings("ignore", category=ConvergenceWarning)
warnings.filterwarnings("error", category=PerfectSeparationWarning)
# import pandas as pd
# warnings.formatwarning(category=ConvergenceWarning)


@dataclass
class Phecode:
    cases: list[str] = field(default_factory=list)
    exclusions: list[str] = field(default_factory=list)


def read_in_cases_and_exclusions(
    counts_file: Path, min_phecode_count: int
) -> dict[str, Phecode]:

    return_dict = {}

    with xopen(counts_file, "r") as counts:
        _ = next(counts)  # skip header line
        for line in counts:
            person_id, phecode, count = line.strip().split(",")
            if int(count) >= min_phecode_count:
                phecode_obj = return_dict.setdefault(
                    phecode, Phecode()
                )  # We can see if there is already a phecode object create for the phecode
                phecode_obj.cases.append(person_id)
            elif (
                int(count) == 1 and min_phecode_count != 1
            ):  # if the min phecode count is 1 then we want to ignore the exclusion criteria
                phecode_obj = return_dict.setdefault(phecode, Phecode())
                phecode_obj.exclusions.append(person_id)

    print(f"Read in {len(return_dict)} phecodes from the counts file: {counts_file}")
    return return_dict


def _format_results(
    results: BinaryResultsWrapper | GLMResultsWrapper,
) -> dict[str, Any]:
    result_dictionary = {}
    # We need to first get the convered status
    # print(int(observation_count) - case_count)
    result_dictionary["converged"] = str(results.converged)

    # generate a table with the beta, stderr, and pvalue for each variable in the model #TODO: refactor
    print("")
    for key, pvalue in results.pvalues.items():

        # variable, beta, stderr, z, pvalue, *_ = line.strip().replace(" ", "").split(",")
        # # We don't need to report the values for the intercept
        if key == "Intercept":
            continue
        elif (
            key == "phecode_status"
        ):  # if the phecode is being used as a predictor then we need to ultimately
            key = key.split("_")[0]

        result_dictionary[f"{key}_pvalue"] = pvalue
    for key, beta in results.params.items():
        # variable, beta, stderr, z, pvalue, *_ = line.strip().replace(" ", "").split(",")
        # # We don't need to report the values for the intercept
        if key == "Intercept":
            continue
        elif key == "phecode_status":
            key = key.split("_")[0]

        result_dictionary[f"{key}_beta"] = beta
    for key, se in results.bse.items():
        # variable, beta, stderr, z, pvalue, *_ = line.strip().replace(" ", "").split(",")
        # # We don't need to report the values for the intercept
        if key == "Intercept":
            continue
        elif key == "phecode_status":
            key = key.split("_")[0]

        result_dictionary[f"{key}_stderr"] = se
        #     result_dictionary[f"{variable}_beta"] = beta
        #     result_dictionary[f"{variable}_stderr"] = stderr

    return result_dictionary


def generate_model_str(
    covar_list: list[str] | None,
    status_col: str,
    flip_predictor_and_outcome: bool = False,
) -> str:
    # build the model string where the phecode status is the
    # outcome and the predictor is our case control status. When the
    # regression code adds the phecode status to the covariate df, it
    # names it phecode_status so we can assume that is the name of the
    # column in the model
    if not flip_predictor_and_outcome:
        analysis_str = f"phecode_status ~ {status_col}"
    else:
        analysis_str = f"{status_col} ~ phecode_status"

    if covar_list:
        analysis_str = f"{analysis_str} + {' + '.join(covar_list)}"

    print(f"model being used: {analysis_str}")
    return analysis_str


@dataclass
class RegressionResults:
    model_type: str
    result: dict[str, Any]
    err: Exception | None


def run_regression(
    model_eq: str,
    covariates: pl.DataFrame,
    regression_model: str,
    max_iteration_count: int,
) -> RegressionResults:
    covariate_df = covariates.to_pandas()
    try:
        if regression_model == "logistic":
            result = smf.logit(model_eq, data=covariate_df).fit(
                disp=0, maxiter=max_iteration_count
            )
        else:
            # run the linear model
            result = smf.glm(
                formula=model_eq, data=covariate_df, family=Gaussian()
            ).fit(disp=0, maxiter=max_iteration_count)

        result = _format_results(result)
        error = None

        # In the case of any exception we are going to just return the error and result will be None
    except Exception as e:
        result = {}
        error = e

    return RegressionResults(model_type=regression_model, result=result, err=error)


def check_err(error_obj: Exception) -> int:
    """look at the exception and decide whether the program can continue or if it needs to fail"""
    # If a perfect separation error occurs then we
    if (
        "Singular matrix" in str(error_obj)
        or "Perfect separation" in str(error_obj)
        or "overflow encountered in exp" in str(error_obj)
    ):
        return 0
    else:
        raise error_obj


def run_phewas(
    phecode_info: tuple,
    return_dictionary: DictProxy,
    covariates: pl.DataFrame,
    analysis_str: str,
    sample_colname: str,
    min_case_count: int,
    max_iteration_threshold: int,
    regression_model: str = "logistic",
) -> None:
    """method to run regress for the phenotype of interest
    Parameters
    ----------
    phecode_info : tuple
        tuple that has information for the regression such as what the phecode name is and who are the cases/controls/and exclusions

    return_dictionary : DictProxy
        Dictionary that is being shared between the different
        python processes.

    covariates : pl.DataFrame
        polars dataframe containing all of the covariates to use
        in the regression analysis

    analysis_str : str
        string for the model that takes the form 'Y=mx+B'

    sample_colname : str,
        column that has the sample ids for the analysis

    min_case_count : int
        minimum number of cases that a phecode has to have to be
        included in the analysis

    max_iteration_threshold : int
        maximum number of iterations for the model to converge.
        This value can be increased if a large number of the
        regressions don't converge

    regression_mode : str
        string indicating whether we are trying to run a
        linear regression model or a logistic regression model

    """

    phecode_name, phecode_obj = phecode_info[0]

    covariates_df = covariates.filter(
        ~pl.col(sample_colname).is_in(phecode_obj.exclusions)
    ).with_columns(
        pl.when(pl.col(sample_colname).is_in(phecode_obj.cases))
        .then(1)
        .otherwise(0)
        .alias("phecode_status")
    )  # Note that anyone here who was not in the counts file (and therefore is
    # not a case or exclusions) will be a control. This means if the individual
    # has no count for any phecode then they will be considered a control

    # This check ensures that the number of cases is >= to the minimum case count.
    # There is already a check for case count when the case and exclusions are read in,
    # but this check fails if there are people in the counts file that are not in the
    # covariate file (pyphewas uses the covariate file to build the cohort). When there
    # are extract individuals in the counts file then the case count can be higher than
    # it would be in the cohort
    if covariates_df.select(pl.col("phecode_status")).sum().item() < min_case_count:
        return
    # When doing the sex stratified analysis there is a chance that there will be no cases
    # despite how there were originally cases. We need to account for that here.
    if len(covariates_df.select(pl.col("phecode_status").value_counts())) == 1:
        return
    # lets get the counts of cases and controls (This is verbose but adapted from the
    # stackoverflow answer: https://stackoverflow.com/questions/78057705/is-there-a-simple-way-to-access-a-value-in-a-polars-struct)
    case_count = (
        covariates_df.select(pl.col("phecode_status").value_counts())
        .filter(pl.col("phecode_status").struct["phecode_status"] == 1)
        .item()["count"]
    )
    control_count = (
        covariates_df.select(pl.col("phecode_status").value_counts())
        .filter(pl.col("phecode_status").struct["phecode_status"] == 0)
        .item()["count"]
    )

    # We can calculate exclusions by subtracting the final shape from the original shape
    exclusion_count = covariates.shape[0] - covariates_df.shape[0]

    # run the regression
    results = run_regression(
        analysis_str, covariates_df, regression_model, max_iteration_threshold
    )

    if results.err is not None:
        # Check if the error is a perfect separation error or if
        # it needs to crash the program
        _ = check_err(results.err)
        print(
            f"Perfect separation encountered for phecode {phecode_name}. There were {case_count} cases and {control_count} controls for the phecode."
        )
        return

    # Lets add the counts to the results dictionary
    results.result["case_count"] = case_count
    results.result["control_count"] = control_count
    results.result["exclusion_count"] = exclusion_count
    return_dictionary[phecode_name] = results.result


def _generate_header(
    status_name: str, covar_list: list[str], predictor_output_flipped: bool
) -> str:

    # if the predictor was split then we want the string to start with
    # predictor_phecode and all the columns with the status_name should
    # just say phecode_*
    if predictor_output_flipped:
        phecode_name_str = "predictor_phecode"
    else:  # otherwise the string starts with phecode and uses the status name we provided
        phecode_name_str = "phecode"

    header_str = f"{phecode_name_str}\tphecode_description\tphecode_category\tcase_count\tcontrol_count\texclusion_count\tconverged"

    header_str += f"\t{status_name}_pvalue"
    header_str += f"\t{status_name}_beta"
    header_str += f"\t{status_name}_stderr"

    for covariate in covar_list:
        header_str += f"\t{covariate}_pvalue"
        header_str += f"\t{covariate}_beta"
        header_str += f"\t{covariate}_stderr"

    return header_str


def _write_to_file(
    output_filehandle: io.TextIOWrapper,
    status_name: str,
    phecode_descriptions: dict[str, str],
    covar_list: list[str],
    results: DictProxy,
    predictor_output_flipped: bool = False,
) -> None:

    header = _generate_header(status_name, covar_list, predictor_output_flipped)

    output_filehandle.write(f"{header}\n")

    for phecode_name, phecode_results in results.items():

        phecode_description, category = phecode_descriptions.get(
            phecode_name, ("N/A", "N/A")
        )

        output_str = f"{phecode_name}\t{phecode_description}\t{category}\t{phecode_results.get('case_count', 'N/A')}\t{phecode_results.get('control_count', 'N/A')}\t{phecode_results.get('exclusion_count', 'N/A')}\t{phecode_results.get('converged', 'N/A')}"

        # lets add all the values for the status to the string first

        output_str += f"\t{phecode_results.get(status_name + '_pvalue', 'N/A')}"
        output_str += f"\t{phecode_results.get(status_name + '_beta', 'N/A')}"
        output_str += f"\t{phecode_results.get(status_name + '_stderr', 'N/A')}"

        for covariate in covar_list:
            output_str += f"\t{phecode_results.get(covariate + '_pvalue', 'N/A')}"
            output_str += f"\t{phecode_results.get(covariate + '_beta', 'N/A')}"
            output_str += f"\t{phecode_results.get(covariate + '_stderr', 'N/A')}"

        output_filehandle.write(f"{output_str}\n")


def read_in_phecode_descriptions(descriptions_filepath: Path | None) -> dict[str, str]:
    description_dict = {}

    if descriptions_filepath:
        with xopen(descriptions_filepath, "r") as desc_filehandle:
            reader = csv.reader(desc_filehandle, delimiter=",", quotechar='"')
            _ = next(reader)
            for row in reader:
                phecode, _, _, phecode_str, _, category, *_ = row

                description_dict[phecode] = (phecode_str, category)
    return description_dict


def restrict_covars_to_specific_sex(
    covar_df: pl.DataFrame, sex_option: str, sex_col: str, males_as_one: bool
) -> pl.DataFrame:

    # lets first set the status for male or female
    if males_as_one:
        male_coding = 1
        female_coding = 0
    else:
        male_coding = 0
        female_coding = 1
    # Now we can filter to either male only or female only

    if sex_option.lower() == "female-only":
        output_df = covar_df.filter(pl.col(sex_col) == female_coding)
    else:
        output_df = covar_df.filter(pl.col(sex_col) == male_coding)

    print(
        f"restricting the cohort to {sex_option}, limited the cohort to {output_df.shape[0]} individuals"
    )

    return output_df


def main() -> None:

    parser = generate_parser()

    args = parser.parse_args()

    # We need to make sure that if the "--run-sex-specific" flag was provided that also the "--sex-col" flag was provided
    if args.run_sex_specific and not args.sex_col and not args.male_as_one:
        parser.error(
            f"Detected a provided value of {args.run_sex_specific} for the '--run-sex-specific' flag, but a value was not provided for the '--sex-col' flag or the '--male-as-one' flag. All three flags are required to run a sex specific analysis."
        )

    # we also need to make sure that the sex columns is not
    # in the list of covariates when we are running sex specific.
    if (  # Because args.covariate_list can be none we need to use short circuiting logic. If
        args.covariate_list  # args.covariate_list == None and we try to use the in operator then the code will raise an error
        and args.sex_col
        in args.covariate_list  # to avoid this we check if the covariate list is not none by checking
        and args.run_sex_specific  # for a truth value first
    ):
        parser.error(
            f"detected that the sex or gender column, {args.sex_col}, was also passed as a covariate. If you are trying to run a sex specific analysis please make sure you remove the sex or gender column from the analysis."
        )

    # getting the programs start time
    start_time = datetime.now()
    # we need to determine the correct path for the phecode descriptions.
    # We can store this values in config file
    if not args.phecode_descriptions and args.phecode_version in [
        "phecodeX",
        "phecode1.2",
        "phecodeX_who",
    ]:
        main_filepath = Path(__file__)
        with open(main_filepath.parent / "config.json", "r") as json_file:
            configs = json.load(json_file)
            try:
                phecode_descriptions = (
                    main_filepath.parent / configs[args.phecode_version]
                )
            except KeyError:
                print(
                    f"The provided phecode version, {args.phecode_version} is not allowed. Please provide a value of either 'phecodeX', 'phecode1.2', or 'phecodeX_who' spelled exactly as shown here"
                )
                sys.exit(1)
    # If the phecode version is none then we need to set the
    # phecode_description variable also to none indicating we
    # will not have any phecode descriptions
    elif not args.phecode_descriptions and args.phecode_version == "None":
        phecode_descriptions = None
    else:
        phecode_descriptions = args.phecode_descriptions

    print(f"{35*'~'}  PheWAS  {35*'~'}")
    print(f"Analysis start time: {start_time}")
    print(
        f"Using the following covariates in the analysis: {', '.join(args.covariate_list if args.covariate_list else '')}"
    )
    print(f"Using {args.cpus} cpus")

    if not args.flip_predictor_and_outcome:
        print(f"Variable of interest column name: {args.status_col}")
    else:
        print(
            f"Using the column, {args.status_col}, as the outcome. Every phecode in the provided counts file, {args.counts}, will be used as a predictor"
        )

    print(
        f"Requiring {args.min_phecode_count} occurences of the PheCode to be considered a case"
    )
    print(
        f"Requiring {args.min_case_count} cases for a phecode to be included in the analysis"
    )
    print(
        f"Using a maximum number of {args.max_iterations} iterations for the logistic regression model"
    )
    print(f"{80*'~'}\n")
    print(f"Loading in the phecode descriptions found here: {phecode_descriptions}")

    descriptions = read_in_phecode_descriptions(phecode_descriptions)

    phecode_cases = read_in_cases_and_exclusions(args.counts, args.min_phecode_count)

    covariates_df = pl.read_csv(args.covariate_file)

    if args.run_sex_specific:
        print("Restricting the covariates file to {args.run_sex_specific}")
        covariates_df = restrict_covars_to_specific_sex(
            covariates_df, args.run_sex_specific, args.sex_col, args.male_as_one
        )

    print("initializing multiprocessing" * (min(args.cpus - 1, 1)))

    item_count = len(phecode_cases.keys())

    # generating the analysis string for the model that will
    # be passed to the regression
    model_str = generate_model_str(
        args.covariate_list, args.status_col, args.flip_predictor_and_outcome
    )

    original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
    with (
        mp.get_context("spawn").Pool(processes=args.cpus) as pool,
        xopen(args.output, "w") as output_file,
        tqdm(total=item_count, desc="phecodes processed") as pbar,
    ):
        signal.signal(signal.SIGINT, original_sigint_handler)
        # We are going to create a manager and a dictionary to use in the regressions
        manager = mp.Manager()
        managed_dict = manager.dict()

        # running the logistic regression for multiple phecodes
        partial_func = partial(
            run_phewas,
            return_dictionary=managed_dict,
            covariates=covariates_df.clone(),
            analysis_str=model_str,
            sample_colname=args.sample_col,
            min_case_count=args.min_case_count,
            max_iteration_threshold=args.max_iterations,
        )
        try:
            for _ in pool.imap(
                partial_func, [(item,) for item in phecode_cases.items()]
            ):
                pbar.update()
        except KeyboardInterrupt:
            print("Detected a keyboard interuption. Ending program now")
            pool.terminate()
            print(f"{30 * '~'}  PheWAS Finished!  {30 * '~'}")
            sys.exit(1)
        else:
            pool.close()
            pool.join()

        # results = run_logit_regression(managed_dict, phecode_cases, covariates_df, args.covariate_list, args.status_col, args.sample_col,args.min_case_count)

        phecodes_tested = len(managed_dict)

        if phecodes_tested == 0:
            print("No phecodes were successfully tested. terminating program")
            sys.exit(1)
        else:
            bonferroni = 0.05 / phecodes_tested
        print(
            f"recommend Bonferroni correction: {bonferroni} or {-np.log10(bonferroni)} on a -log10 scale"
        )

        print(f"Writing the results of the PheWAS to the file: {args.output}")

        # if we are using the phecodes as the predictor then
        # we want out non covariate columns to just be "phecode_*"
        if args.flip_predictor_and_outcome:
            status_col = "phecode"
        # If we are using our case/control file as predictors then we will use
        # the status column of interest that we said
        else:
            status_col = args.status_col

        _write_to_file(
            output_file,
            status_col,
            descriptions,
            args.covariate_list,
            managed_dict,
            args.flip_predictor_and_outcome,
        )

    end_time = datetime.now()
    print(f"program finished at {end_time}")
    print(f"total runtime: {end_time - start_time}")
    print(f"{30 * '~'}  PheWAS Finished!  {30 * '~'}")


if __name__ == "__main__":
    main()



import argparse
from rich_argparse import RichHelpFormatter
from importlib.metadata import version
from pathlib import Path


def generate_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="CLI tool used to run a PheWAS.",
        formatter_class=RichHelpFormatter,
        epilog="""
    Minimal Examples:
        %(prog)s  --counts counts.csv --covariate-file covariates.csv --min-phecode-count 2 --covariate-list COVAR_1 COVAR_2 COVAR_3 --status-col status_colname --sample-col sample_colname --min-case-count 100
    """,
    )

    parser.add_argument(
        "--counts", type=Path, help="counts of each phecode that someone has"
    )

    parser.add_argument(
        "--covariate-file",
        type=Path,
        help="comma separated file that has the covariates for each individual in the cohort. There should be a column for the individual ids, the variable of interest, and then covariates. Even if you are running this without covariates then there should still be the individual ids and variable of interest columns.",
    )

    parser.add_argument(
        "--min-phecode-count",
        type=int,
        default=2,
        help="minumum number of times the person needs the PheCode for them to count as a case. Default %(default)s",
    )

    parser.add_argument(
        "--min-case-count",
        type=int,
        default=20,
        help="minumum number of cases for a phecode to have to be included in the analysis. Default %(default)s",
    )

    parser.add_argument(
        "--covariate-list",
        nargs="*",
        help="list of covariates to use in the analysis. These variables have to be spelled exactly like how they are in the covariates file. If there is a sex or gender column in the file, it is assuming that the values in the column are either 0 or 1",
    )

    parser.add_argument(
        "--status-col",
        type=str,
        default="status",
        help="column that has the variable of interest. Default: %(default)s",
    )

    parser.add_argument(
        "--sample-col",
        type=str,
        default="person_id",
        help="Column name for the column in the covariates file that has the individual ids. Default: %(default)s",
    )

    parser.add_argument(
        "--output",
        "-o",
        type=Path,
        default=Path("test_output.txt"),
        help="output_file that the results for the phewas will be written to. If the filename ends in .gz then the file will be gzipped. Otherwise it will just be a tab separated text file. Default: %(default)s",
    )

    parser.add_argument(
        "--phecode-descriptions",
        type=Path,
        help="Comma separate file that list the phecode and the phecode description. The phecode is expected to be the first colummn while the phecode description is expected to be the fourth column.",
    )

    parser.add_argument(
        "--cpus",
        type=int,
        default=1,
        help="number of cpu processes to use in the program. This will be passed to the multiprocess.Pool. Default: %(default)s",
    )

    parser.add_argument(
        "--max-iterations",
        type=int,
        default=200,
        help="Maximum number of iterations for the model to go through before it converges. If the model doesn't converge after this value then a ConvergenceWarning will be thrown. Default: %(default)s",
    )

    parser.add_argument(
        "--phecode-version",
        type=str,
        default="None",
        choices=["phecodeX", "phecode1.2", "phecodeX_who", "None"],
        help="What version of phecodes to use for the analysis. The options have to be spelled the same as they are here. Allowed choices: '%(choices)s'",
    )

    parser.add_argument(
        "--model",
        type=str,
        default="logistic",
        required=False,
        choices=["linear", "logistic"],
        help="Whether to run a linear model or a logistic model for the regression. (default: %(default)s)",
    )

    parser.add_argument(
        "--flip-predictor-and-outcome",
        default=False,
        help="Optional flag to control the behavior of what is the predictor and what is the outcome. By default, our model assumes the outcome is the phecode status and the predictor is or provided case/control status. If this flag is provided then the outcome becomes our case/control status and the predictor will be the phecode status %(default)s",
        action="store_true",
    )

    parser.add_argument(
        "--run-sex-specific",
        type=str,
        choices=["male-only", "female-only"],
        help="Flag that allows the user to run the analysis sex specific. By default the PheWAS will be sex agonistic, allowing the user to adjust for sex using covariates. If the user provides a value then the program will filter to either males or females based on the value provided. %(choices)s",
    )

    parser.add_argument(
        "--male-as-one",
        default=True,
        type=bool,
        help="Flag indicating whether males are labeled as 1 and females are labeled as 0 or vice versa. %(default)s",
    )

    parser.add_argument(
        "--sex-col",
        type=str,
        help="Column name of the column in the covariates file that contains information about sex for each individual. This argument only needs to be provided if the user is using the '--run-sex-specific' flag. If this argument is passed and the '--run-sex-specific' flag is not provided then the program will ignore this flag. If there are people with missing sex values in the covariates file then they will be implicited filtered out of the analysis",
    )

    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s: {version('pyphewas-package')}",
    )

    return parser



from rich_argparse import RichHelpFormatter
import numpy as np
import argparse
from pathlib import Path
import polars as pl
import matplotlib.pyplot as plt
import seaborn as sns
from dataclasses import dataclass


@dataclass
class FormattedDf:
    dataframe: pl.DataFrame
    infinity_threshold: float


def determine_color_palatte(category_count: int) -> list[tuple]:
    """return the color palette based off of how many different phecode
    categories there are

    Parameters
    ----------
    category_count : int
        number of unique phecode categories in the provided dataframe file

    Returns
    -------
    list[tuple]
        returns a list different colors to represent each phecodee category"""
    if category_count <= 10:
        distinct_palette = sns.color_palette("tab10", n_colors=category_count)
    else:
        distinct_palette = sns.color_palette("tab20", n_colors=category_count)

    return distinct_palette


def format_df(df: pl.DataFrame, pval_colname: str, beta_colname: str) -> FormattedDf:
    """format the provided data in the following ways: 1) only keep converged
    values, 2) replace pvalues that were too significant and rounded to zero
    to be 1.02 * the most significant pvalue,3) create a column that indicates
    where the beta was negative or positive, 4) sort by phecode_category and
    phecode_description columns, 5) and a row index column that will be used to
    plot things on the x axis

    Parameters
    ----------
    df : pl.DataFrame
        dataframe that has the results from running the phewas.

    pval_colname : str
        name of the column that has the pvalue of interest

    beta_colname : str
        name of the column that has the beta for the vafriable of interest. We
        will use this column to create another column of whether the betas are
        negative or not

    Returns
    -------
    pl.DataFrame
        returns the dataframe frame with the aforementioned transformations
    """
    # Ensure that the provided df has the correct values
    check_df_columns(df, pval_colname, beta_colname)

    df = df.filter(pl.col("converged")).with_columns(
        -np.log10(pl.col(pval_colname)).alias("neg_log10_p")
    )

    max_finite_val = (
        df.filter(pl.col("neg_log10_p").is_finite())
        .select(pl.col("neg_log10_p").max())
        .item()
        * 1.02
    )

    print(f"replacing infinite values in the dataframe with {max_finite_val}")

    df = (
        df.with_columns(
            pl.when(~pl.col("neg_log10_p").is_finite())
            .then(max_finite_val)
            .otherwise(pl.col("neg_log10_p"))
        )
        .with_columns(
            pl.when(pl.col(beta_colname) > 0)
            .then(pl.lit("Positive"))
            .otherwise(pl.lit("Negative"))
            .alias("direction")
        )
        .sort(["phecode_category", "phecode_description"])
        .with_row_index(name="index")
    )

    return FormattedDf(df, max_finite_val)


def generate_manhatten(
    formatted_results: FormattedDf,
    output_filename: Path | str,
    significance_threshold: float,
    dpi: int = 300,
    color_palette: list[tuple] | None = None,
) -> None:

    df = formatted_results.dataframe

    plt.figure(figsize=(16, 9))

    if color_palette is None:
        color_palette = determine_color_palatte(len(df["phecode_category"].unique()))

    # We need to format the dataframe
    markers = {"Positive": "^", "Negative": "v"}

    plot_data = df.to_pandas()

    sns.scatterplot(
        data=plot_data,
        x="index",
        y="neg_log10_p",
        hue="phecode_category",
        palette=color_palette,
        style="direction",
        markers=markers,
        s=80,
        alpha=1.0,  # Set alpha to 1.0 for maximum color saturation
        edgecolor="black",  # Black edge helps separate the colors further
        linewidth=0.5,
    )

    # Threshold lines
    plt.axhline(
        y=significance_threshold,
        color="#ff0000",
        linestyle="--",
        linewidth=1.5,
        label="Bonferroni",
    )

    plt.axhline(
        y=formatted_results.infinity_threshold,
        color="#1e90ff",
        linestyle="--",
        linewidth=1.5,
        label="infinity_threshold",
    )

    # generate the differnt category labels so everything aligns correctly on the x-axis
    label_df = (
        df.group_by("phecode_category")
        .agg(pl.col("index").median().alias("pos"))
        .sort("pos")
    )
    plt.xticks(
        label_df["pos"].to_list(),
        label_df["phecode_category"].to_list(),
        rotation=45,
        ha="right",
        fontsize=10,
    )

    plt.xlabel("Phecode Categories", fontsize=14)
    plt.ylabel(r"$-\log_{10}(P)$", fontsize=14)
    plt.title("PheWAS Results", fontsize=14)
    plt.legend(
        bbox_to_anchor=(1.01, 1),
        loc="upper left",
        borderaxespad=0,
        title="Category / Direction",
    )
    plt.tight_layout()
    plt.savefig(output_filename, dpi=dpi, bbox_inches="tight")


def check_df_columns(df: pl.DataFrame, pval_col: str, beta_col: str) -> None:
    """check if the expected columns are in the dataframe. If the expected columns
    are not then then we need to terminate the program

    Parameters
    ----------
    df : pl.DataFrame
        dataframe that has the results from the phewas.

    pval_col : str
        column that has the pvalues for the variable of interest

    beta_col : str
        column that has the betas for the variable of interest

    Raises
    ------
    ValueError
        if any of the expected columns are not found then we will raise a value error
        telling the user what the expected columns are
    """
    cols_to_search_for = [
        "phecode_category",
        "phecode_description",
        "converged",
        pval_col,
        beta_col,
    ]

    cols_present = [col not in df.columns for col in cols_to_search_for]

    if any(cols_present):
        err_msg = (
            "Did not find all the expected columns within the provided dataframe. Please make sure the following columns are in the dataframe: "
            + ", ".join(cols_present)
        )
        raise ValueError(err_msg)


def main() -> None:

    parser = argparse.ArgumentParser(
        description="CLI too to generate the manhattan plot after running the PheWAS",
        formatter_class=RichHelpFormatter,
        epilog="""
    Minimal Examples:
        %(prog)s  --input-file phewas_results.txt 
    """,
    )

    parser.add_argument(
        "--input-file",
        "-i",
        help="input file with the reusults from the PheWAS. This file should be tab separated and have the columns 'phecode_category', 'phecode_description', 'coverged'.",
    )

    parser.add_argument(
        "--output-file",
        "-o",
        type=Path,
        help="filepath to output the manhattan plot to. (default: %(default)s)",
        default=Path("test.png"),
    )

    parser.add_argument(
        "--pval-col",
        type=str,
        help="name of the column that has the pvalues for the variable of interest. (default: %(default)s)",
        default="pval",
    )

    parser.add_argument(
        "--beta-col",
        type=str,
        help="name of the column that has betas for the variables of interest. (default: %(default)s)",
        default="beta",
    )

    parser.add_argument(
        "--dpi",
        type=int,
        default=300,
        help="quality of the image to output. (default: %(default)s)",
    )

    args = parser.parse_args()

    # read in the dataframe
    df = pl.read_csv(args.input_file, separator="\t")

    significance_threshold = 0.05 / df.shape[0]
    formatted_df_results = format_df(df, args.pval_col, args.beta_col)

    generate_manhatten(
        formatted_df_results, args.output_file, significance_threshold, dpi=args.dpi
    )


if __name__ == "__main__":
    main()



import sys
import subprocess
from pathlib import Path

# Define paths to input files
TESTS_DIR = Path(__file__).parent
INPUTS_DIR = TESTS_DIR / "inputs"

LOGISTIC_COUNTS = (
    INPUTS_DIR / "phecodes_file_10000_samples_binary_predictor_30_phecodes.txt"
)
LOGISTIC_COVARIATES = (
    INPUTS_DIR / "covariates_file_10000_samples_binary_predictor_30_phecodes.txt"
)

LINEAR_COUNTS = (
    INPUTS_DIR / "phecodes_file_10000_samples_dosage_predictor_30_phecodes.txt"
)
LINEAR_COVARIATES = (
    INPUTS_DIR / "covariates_file_10000_samples_dosage_predictor_30_phecodes.txt"
)

# Path to the main script
SCRIPT_PATH = Path("src/pyphewas/run_PheWAS.py")


def test_logistic_regression(tmp_path):
    output_file = tmp_path / "logistic_output.txt"

    cmd = [
        sys.executable,
        str(SCRIPT_PATH),
        "--counts",
        str(LOGISTIC_COUNTS),
        "--covariate-file",
        str(LOGISTIC_COVARIATES),
        "--min-phecode-count",
        "2",
        "--min-case-count",
        "20",
        "--covariate-list",
        "age",
        "sex",
        "--status-col",
        "predictor",
        "--sample-col",
        "id",
        "--phecode-version",
        "None",
        "--model",
        "logistic",
        "--output",
        str(output_file),
        "--cpus",
        "2",
    ]

    # Run the command
    result = subprocess.run(cmd, capture_output=True, text=True)

    # Check that the command ran successfully
    assert result.returncode == 0, f"Command failed with stderr: {result.stderr}"

    # Check that the output file was created
    assert output_file.exists()

    assert output_file.stat().st_size > 0


def test_linear_regression(tmp_path):
    output_file = tmp_path / "linear_output.txt"

    cmd = [
        sys.executable,
        str(SCRIPT_PATH),
        "--counts",
        str(LINEAR_COUNTS),
        "--covariate-file",
        str(LINEAR_COVARIATES),
        "--min-phecode-count",
        "2",
        "--min-case-count",
        "20",
        "--covariate-list",
        "age",
        "sex",
        "--status-col",
        "predictor",
        "--sample-col",
        "id",
        "--phecode-version",
        "None",
        "--model",
        "linear",
        "--output",
        str(output_file),
        "--cpus",
        "2",
    ]

    # Run the command
    result = subprocess.run(cmd, capture_output=True, text=True)

    # Check that the command ran successfully
    assert result.returncode == 0, f"Command failed with stderr: {result.stderr}"

    # Check that the output file was created
    assert output_file.exists()

    assert output_file.stat().st_size > 0
